<?php
namespace nulib\file\cache;

use nulib\file\SharedFile;
use nulib\os\path;
use nulib\php\time\DateTime;
use nulib\php\time\Delay;
use nulib\str;

class CacheFile extends SharedFile {
  /** @var string|int durée de vie par défaut des données mises en cache  */
  const DURATION = "1D"; // jusqu'au lendemain

  function __construct($file, ?array $params=null) {
    if ($file === null) {
      $rand = bin2hex(random_bytes(8));
      $file = sys_get_temp_dir()."/$rand";
    }
    $file = path::ensure_ext($file, ".metadata.cache", ".cache");
    $this->basedir = path::dirname($file);
    $basename = path::filename($file);
    $this->basename = str::without_suffix(path::ext($basename), $basename);
    $this->duration = Delay::with($params["duration"] ?? static::DURATION);
    $this->overrideDuration = $params["override_duration"] ?? false;
    $this->cacheNull = $params["cache_null"] ?? false;
    $this->datafiles = [];
    parent::__construct($file);
  }

  /** @var string répertoire de base des fichiers de cache */
  protected string $basedir;

  /** @var string nom de base des fichiers de cache */
  protected string $basename;

  protected Delay $duration;

  protected bool $overrideDuration;

  protected bool $cacheNull;

  protected array $datafiles;

  /**
   * vérifier si le fichier est valide. s'il est invalide, il faut le recréer.
   *
   * on assume que le fichier existe, vu qu'il a été ouvert en c+b
   */
  protected function isValid(): bool {
    # considèrer que le fichier est invalide s'il est de taille nulle
    return $this->getSize() > 0;
  }

  /** charger les données. le fichier a déjà été verrouillé en lecture */
  protected function loadData(): ?array {
    $this->rewind();
    [
      "start" => $start,
      "duration" => $duration,
      "datafiles" => $datafiles,
    ] = $this->unserialize(null, false, true);
    if ($this->overrideDuration) {
      $duration = Delay::with($this->duration, $start);
    }
    return [
      "start" => $start,
      "duration" => $duration,
      "datafiles" => $datafiles,
    ];
  }

  /** tester s'il faut mettre les données à jour. le fichier a déjà été verrouillé en lecture */
  protected function shouldUpdate(bool $noCache=false): bool {
    if ($this->isValid()) {
      /** @var Delay $duration */
      ["duration" => $duration,
      ] = $this->loadData();
      $expired = $duration->isElapsed();
    } else {
      $expired = false;
      $noCache = true;
    }
    return $noCache || $expired;
  }

  /** sauvegarder les données. le fichier a déjà été verrouillé en écriture */
  protected function saveData(?DateTime $start=null, ?Delay $duration=null): void {
    $duration ??= $this->duration;
    if ($start === null) {
      $start = new DateTime();
      $duration = Delay::with($duration, $start);
    }
    $this->ftruncate();
    $this->serialize([
      "start" => $start,
      "duration" => $duration,
      "datafiles" => $this->datafiles,
    ], false, true);
  }

  /** tester si $value peut être mis en cache */
  function shouldCache($value): bool {
    return $this->cacheNull || $value !== null;
  }

  /** obtenir les informations sur le fichier */
  function getInfos(): array {
    $this->lockRead();
    try {
      if (!$this->isValid()) {
        return ["valid" => false];
      }
      /**
       * @var DateTime $start
       * @var Delay $duration
       */
      [
        "start" => $start,
        "duration" => $duration,
        "datafiles" => $datafiles,
      ] = $this->loadData();
      return [
        "valid" => true,
        "size" => $this->getSize(),
        "start" => $start,
        "duration" => strval($duration),
        "date_start" => $start->format(),
        "date_end" => $duration->getDest()->format(),
        "datafiles" => $datafiles,
      ];
    } finally {
      $this->unlock();
    }
  }

  const UPDATE_SUB = -1, UPDATE_SET = 0, UPDATE_ADD = 1;

  /** mettre à jour la durée de validité du fichier */
  function updateDuration($nduration, int $action=1): void {
    $this->lockRead();
    try {
      if (!$this->isValid()) return;
      $this->lockWrite();
      /**
       * @var DateTime $tstart
       * @var Delay $duration
       */
      [
        "start" => $start,
        "duration" => $duration,
      ] = $this->loadData();
      if ($action < 0) $duration->subDuration($nduration);
      elseif ($action > 0) $duration->addDuration($nduration);
      $this->saveData($start, $duration);
    } finally {
      $this->unlock();
    }
  }

  /** supprimer les fichiers s'ils ont expiré */
  function deleteExpired(bool $force=false): bool {
    $this->lockRead();
    try {
      if ($force || $this->shouldUpdate()) {
        $this->lockWrite();
        @unlink($this->file);
        $basedir = $this->basedir;
        foreach ($this->datafiles as $datafile) {
          @unlink(path::join($basedir, $datafile));
        }
        return true;
      }
    } finally {
      $this->unlock();
    }
    return false;
  }
}
